////////////////////////////////////////////////////////////////////////////
// 2015 UTC Buildings & Industrial Systems. All rights reserved.
// This software contains information which represents trade secrets of UTC Buildings & Industrial Systems and may not be copied or disclosed
// to others except as provided in the license with UTC Buildings & Industrial Systems.

// Code in this module is based on the demo application, testRTSPCLient, copied from Live555 website.
// The code underwent significant modifications to suit UTC needs.
// For more details about testRTSPClient demo application and OpenRTSP, see http://www.live555.com/openRTSP/
////////////////////////////////////////////////////////////////////////////
// RTSP Async Adapter
//    Adapter to use new asynchronous RTSP interface of the new Live555
//    All streaming clients using RTSP, like session_lnvr and session_live555
//    are expected to use this adapter.
////////////////////////////////////////////////////////////////////////////

#include "RTSPAsyncAdapter.h"
#include "../../../common/utils/dbg_out.h"
#include "../../../common/utils/logging.h"
#include <time.h>

// TODO: use dbg() or log_internal_write()
#if _DEBUG && 0 // Suppress logs for now
#define LOCAL_DEBUG_0(s,l) dbg("*** %s[%d][%d][%d]:\n",s,l,GetCurrentThreadId(),(DWORD)time(NULL))
#define LOCAL_DEBUG_1S(s,l, s1) dbg("*** %s[%d][%d][%d]: %s\n",s,l,GetCurrentThreadId(),(DWORD)time(NULL), s1)
#define LOCAL_DEBUG_2S(s,l, s1,s2) dbg("*** %s[%d][%d][%d]: %s, %s\n",s,l,GetCurrentThreadId(),(DWORD)time(NULL), s1,s2)
#define LOCAL_DEBUG_SL(s,l, s1,l1) dbg("*** %s[%d][%d][%d]: %s, %d\n",s,l,GetCurrentThreadId(),(DWORD)time(NULL), s1,l1)
#define LOCAL_DEBUG_1L(s,l, l1) dbg("*** %s[%d][%d][%d]: %d\n",s,l,GetCurrentThreadId(),(DWORD)time(NULL), (DWORD)l1)
#define LOCAL_DEBUG_2L(s,l, l1, l2) dbg("*** %s[%d][%d][%d]: %d, %d\n",s,l,GetCurrentThreadId(),(DWORD)time(NULL), (DWORD)l1, (DWORD)l2)
#define LOCAL_DEBUG_3L(s,l, l1, l2, l3) dbg("*** %s[%d][%d][%d]: %d, %d, %d\n",s,l,GetCurrentThreadId(),(DWORD)time(NULL), (DWORD)l1, (DWORD)l2, (DWORD)l3)
#else
#define LOCAL_DEBUG_0(s,l)
#define LOCAL_DEBUG_1S(s,l, s1)
#define LOCAL_DEBUG_2S(s,l, s1, s2)
#define LOCAL_DEBUG_SL(s,l, s1,l1)
#define LOCAL_DEBUG_1L(s,l, l1)
#define LOCAL_DEBUG_2L(s,l, l1, l2)
#define LOCAL_DEBUG_3L(s,l, l1, l2, l3)
#endif


////////////////////////////////////////////////////////////////////////////
// definitions for macros and globals
////////////////////////////////////////////////////////////////////////////
#define RTSP_CLIENT_VERBOSITY_LEVEL 1 // by default, print verbose output from each "RTSPClient"
#define RAA_E_SDP_FAILED                  "%s: Failed to get SDP Description\n"
#define RAA_E_MEDIA_SESS_CREATION_FAILURE "%s: Failed to create a MediaSession object from the SDP description\n"
#define RAA_E_MEDIA_SESS_LIST_EMPTY       "%s: Session has no media sessions\n"
#define RAA_E_INITIATE_FAILED             "%s: Subsession initiation failed (%s)\n"
#define RAA_E_SUBSESS_SETUP_FAILED        "%s: Subsession setup failed (%s)\n"
#define RAA_E_SINK_CREATION_FAILED        "%s: Media sink creation failed (%s)\n"
#define RAA_E_PLAY_SESS_FAILED            "%s: PLAY on session failed (%s)\n"
#define RAA_I_PLAY_DURATION_DEFINED       "%s: PLAY on session duration = %d sec\n"
#define RAA_I_RTCP_BYE_RECEIVED           "%s: RTCP BYE received\n"
#define RAA_E_RTSP_CLIENT_CREATION_FAILED "RTSP Client creation failed\n"

////////////////////////////////////////////////////////////////////////////
// RTSP 'response handlers'
////////////////////////////////////////////////////////////////////////////
void postDESCRIBE(RTSPClient* pClient, int rc, char* pSDP);
void postSETUP(RTSPClient* pClient, int rc, char* rcStr);
void postPLAY(RTSPClient* pClient, int rc, char* rcStr);

////////////////////////////////////////////////////////////////////////////
// Other event handler functions
////////////////////////////////////////////////////////////////////////////
void postSubsessionPLAY(void* cData);
void subsessionByeHandler(void* cData);

// Used to iterate through each stream's 'subsessions', setting up each one:
void setupNextSubsession(RTSPClient* pClient, MediaSubsession *pSubsession);

void shutdownStream(RTSPClient* pClient, int ec = 1);
static void beginStatsMeasurement(void* cData, MediaSession* session);

////////////////////////////////////////////////////////////////////////////
// Code
////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////
// Implementation of the RTSP 'response handlers':
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// postDESCRIBE
////////////////////////////////////////////////////////////////////////////
void postDESCRIBE(RTSPClient* pClient, int rc, char* pSDP)
{
    UsageEnvironment& env = pClient->envir(); // alias
    StreamClientState* pScs = &((RTSPAsyncAdapter*)pClient)->scs; // alias
    RAA_PARAMS* pParams = &((RTSPAsyncAdapter*)pClient)->params;

    if (rc != 0)
    {
        // TODO: Print rerult string
        LOCAL_DEBUG_2S(__FUNCTION__,__LINE__,"DESCRIBE failed: ", pClient->url());
        log_internal_write(L_WARN, __FILE__, __LINE__, RAA_E_SDP_FAILED, pClient->url());
        if (pParams->sdpErrorCB)
        {
            void *p = (void*)pParams;
            pParams->sdpErrorCB(pClient, rc, p);
        }
        shutdownStream(pClient);
        return;
    }
    if (pParams->descCB)
        pParams->descCB(pClient, pParams->cData, rc, pSDP);
    char* sdpDescription = pSDP;
    LOCAL_DEBUG_2S(__FUNCTION__,__LINE__,pParams->sessionMediumName,sdpDescription);

    // Create a media session object from this SDP description:
    pScs->session = MediaSession::createNew(env, (const char*)sdpDescription);
    delete[] sdpDescription;
    if (pScs->session == NULL)
    {
        LOCAL_DEBUG_2S(__FUNCTION__,__LINE__,"Mediasession creation failed: ", pClient->url());
        log_internal_write(L_WARN, __FILE__, __LINE__, RAA_E_MEDIA_SESS_CREATION_FAILURE, pClient->url());
        shutdownStream(pClient);
        return;
    } else if (!pScs->session->hasSubsessions())
    {
        log_internal_write(L_WARN, __FILE__, __LINE__, RAA_E_MEDIA_SESS_LIST_EMPTY, pClient->url());
        shutdownStream(pClient);
        return;
    }

    // Then, create and set up our data source objects for the session.  We do this by iterating over the session's 'subsessions',
    // calling "MediaSubsession::initiate()", and then sending a RTSP "SETUP" command, on each one.
    // (Each 'subsession' will have its own data source.)
    MediaSubsessionIterator iter(*pScs->session);
    MediaSubsession *pSubsession;
    while ((pSubsession = iter.next()) != NULL)
    {
        if (stricmp(pSubsession->mediumName(), pParams->sessionMediumName) == 0)
        {
            setupNextSubsession(pClient, pSubsession);
            // Setup download Speed at session and subsession level
            pClient->setSpeed( *pScs->session, pParams->playbackspeed );
            return;
        }
    }
    setupNextSubsession(pClient, NULL);
}

////////////////////////////////////////////////////////////////////////////
// postDESCRIBE
////////////////////////////////////////////////////////////////////////////
void setupNextSubsession(RTSPClient* pClient, MediaSubsession *pSubsession)
{
    UsageEnvironment& env = pClient->envir(); // alias
    StreamClientState* pScs = &((RTSPAsyncAdapter*)pClient)->scs; // alias
    RAA_PARAMS* pParams = &((RTSPAsyncAdapter*)pClient)->params;
    bool proceed = false;
  
    if (pSubsession != NULL)
    {
        pSubsession->setClientPortNum(0);
        if (!pSubsession->initiate())
        {
            LOCAL_DEBUG_2S(__FUNCTION__,__LINE__,"Receiver creation failed", pSubsession->mediumName());
            log_internal_write(L_WARN, __FILE__, __LINE__, RAA_E_MEDIA_SESS_CREATION_FAILURE, pClient->url(), env.getResultMsg());
            shutdownStream(pClient);
            return;
        }
        else
        {
            LOCAL_DEBUG_2S(__FUNCTION__,__LINE__,"Receiver created", pSubsession->mediumName());
            int payloadType = RAA_PAYLOAD_TYPE_UNKNOWN;
            if (pParams->metaCB)
            {
                std::string codec = pSubsession->codecName();
                LOCAL_DEBUG_1S(__FUNCTION__,__LINE__,codec.c_str());
                if (( codec.find("MP4V-ES") != std::string::npos ) ||
                    ( codec.find("MPEG4") != std::string::npos )   ||
                    ( codec.find("H264") != std::string::npos )    ||
                    ( codec.find("MJPEG") != std::string::npos )   ||
                    ( codec.find("JPEG") != std::string::npos )    ||
                    ( codec.find("YV12") != std::string::npos ))
                {
                    payloadType = RAA_PAYLOAD_TYPE_VIDEO;
                    if ( codec.find("H264") != std::string::npos )
                    {
                        pParams->frameBuf += 3;
                        pParams->frameSize -= 3;
                    }
                }
                else if (( codec.find("PCMU") != std::string::npos ) ||
                        ( codec.find("G711") != std::string::npos ) ||
                        ( codec.find("G726") != std::string::npos ))
                {
                    payloadType = RAA_PAYLOAD_TYPE_AUDIO;
                }
                if (pParams->payloadType == payloadType)
                {
                    LOCAL_DEBUG_2S(__FUNCTION__,__LINE__,"Invoke Metacallback", pSubsession->mediumName());
                    pParams->metaCB(pClient, pParams->cData, 0, (char*)pSubsession->codecName(), (char*)pSubsession->fmtp_spropparametersets());
                    proceed = true;
                } else {
                    LOCAL_DEBUG_2S(__FUNCTION__,__LINE__,"Metacallback not invoked", pSubsession->mediumName());
                }
            }

            if (proceed)
            {
                if (pScs->iter == NULL)
                    pScs->iter = new MediaSubsessionIterator(*pScs->session);
                while ((pScs->subsession = pScs->iter->next()) != NULL)
                {
                    if (stricmp(pScs->subsession->mediumName(), pParams->sessionMediumName) != 0)
                        continue;
                    LOCAL_DEBUG_2S(__FUNCTION__,__LINE__,"Sending SETUP: ", pScs->subsession->mediumName());
                    // Continue setting up this subsession, by sending a RTSP "SETUP" command:
                    // TODO: TCP tunneling & multicast
                    bool tcp = (pParams->netStats.m_strmFlags & SSRM_FL_STREAM_PROTOCOL_TCP) ? true : false;
                    pClient->sendSetupCommand(*pScs->subsession, postSETUP, false, tcp);
                    return;
                }
                if (pScs->iter != NULL)
                {
                    delete pScs->iter;
                    pScs->iter = NULL;
                }
            }
            else
            {
                LOCAL_DEBUG_2S(__FUNCTION__,__LINE__,"Invalid codec -skip sending SETUP: ", pParams->sessionMediumName);
                log_internal_write(L_WARN, __FILE__, __LINE__, RAA_E_MEDIA_SESS_CREATION_FAILURE, pClient->url(), env.getResultMsg());
                shutdownStream(pClient);
            }
        }
    }
    else
        LOCAL_DEBUG_2S(__FUNCTION__,__LINE__,"End of subsessions",pParams->sessionMediumName);
}

////////////////////////////////////////////////////////////////////////////
// postSETUP
////////////////////////////////////////////////////////////////////////////
void postSETUP(RTSPClient* pClient, int rc, char* rcStr)
{
    UsageEnvironment& env = pClient->envir(); // alias
    RAA_PARAMS* pParams = &((RTSPAsyncAdapter*)pClient)->params;
    StreamClientState* pScs = &((RTSPAsyncAdapter*)pClient)->scs; // alias
    do
    {
        if (rc != 0)
        {
            LOCAL_DEBUG_2S(__FUNCTION__,__LINE__,"Error in SETUP: ", pScs->subsession->mediumName());
            log_internal_write(L_WARN, __FILE__, __LINE__, RAA_E_SUBSESS_SETUP_FAILED, pClient->url(), env.getResultMsg());
            shutdownStream(pClient);
            return;
        }

        // Having successfully setup the subsession, create a data sink for it, and call "startPlaying()" on it.
        // (This will prepare the data sink to receive data; the actual flow of data from the client won't start happening until later,
        // after we've sent a RTSP "PLAY" command.)

        LOCAL_DEBUG_2S(__FUNCTION__,__LINE__,"postSETUP: ", pScs->subsession->mediumName());
        pScs->subsession->sink = RAA_Sink::createNew(env, *pScs->subsession,
                                                    (unsigned char*)((RTSPAsyncAdapter*)pClient)->params.frameBuf,
                                                    ((RTSPAsyncAdapter*)pClient)->params.frameSize,
                                                    ((RTSPAsyncAdapter*)pClient)->params.frameCB,
                                                    ((RTSPAsyncAdapter*)pClient)->params.cData,
                                                    pClient->url());

        // perhaps use your own custom "MediaSink" subclass instead
        if (pScs->subsession->sink == NULL)
        {
            log_internal_write(L_WARN, __FILE__, __LINE__, RAA_E_SINK_CREATION_FAILED, pClient->url(), env.getResultMsg());
            shutdownStream(pClient);
            return;
        }

        pScs->subsession->miscPtr = pClient; // a hack to let subsession handle functions get the "RTSPClient" from the subsession 
        pScs->subsession->sink->startPlaying(*(pScs->subsession->readSource()),
        postSubsessionPLAY, pScs->subsession);

        // Also set a handler to be called if a RTCP "BYE" arrives for this subsession:
        if (pScs->subsession->rtcpInstance() != NULL)
        {
            pScs->subsession->rtcpInstance()->setByeHandler(subsessionByeHandler, pScs->subsession);
        }
    } while (0);

    // We've finished setting up all of the subsessions.  Now, send a RTSP "PLAY" command to start the streaming:
    pScs->duration = pScs->session->playEndTime() - pScs->session->playStartTime();
    LOCAL_DEBUG_2S(__FUNCTION__,__LINE__,"Sending PLAY: ", pScs->subsession->mediumName());
    pClient->sendPlayCommand(*pScs->session, postPLAY);
}

////////////////////////////////////////////////////////////////////////////
// postPLAY
////////////////////////////////////////////////////////////////////////////
void postPLAY(RTSPClient* pClient, int rc, char* rs)
{
    UsageEnvironment& env = pClient->envir(); // alias
    RAA_PARAMS* pParams = &((RTSPAsyncAdapter*)pClient)->params;
    StreamClientState* pScs = &((RTSPAsyncAdapter*)pClient)->scs; // alias

    if (rc != 0)
    {
        LOCAL_DEBUG_2S(__FUNCTION__,__LINE__,"Error in PLAY", pScs->subsession->mediumName());
        log_internal_write(L_WARN, __FILE__, __LINE__, RAA_E_PLAY_SESS_FAILED, pClient->url(), env.getResultMsg());
        delete[] rs;
        shutdownStream(pClient);
        return;
    }
    if (pScs->duration > 0)
    {
        LOCAL_DEBUG_2S(__FUNCTION__,__LINE__,"Duration not expected", pScs->subsession->mediumName());
        log_internal_write(L_INFO, __FILE__, __LINE__, RAA_I_PLAY_DURATION_DEFINED, pClient->url(), pScs->duration);
    }
    delete[] rs;
    beginStatsMeasurement(&((RTSPAsyncAdapter*)pClient)->params, pScs->session);

    LOCAL_DEBUG_2S(__FUNCTION__,__LINE__,"Receiving stream data", pParams->sessionMediumName);
}


////////////////////////////////////////////////////////////////////////////
// Implementation of the other event handlers:
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// postSubsessionPLAY: Called when a stream's subsession (e.g., audio or video substream) ends
////////////////////////////////////////////////////////////////////////////
void postSubsessionPLAY(void* clientData)
{
    LOCAL_DEBUG_0(__FUNCTION__,__LINE__);
    MediaSubsession* subsession = (MediaSubsession*)clientData;
    RTSPClient* pClient = (RTSPClient*)(subsession->miscPtr);

    // Begin by closing this subsession's stream:
    Medium::close(subsession->sink);
    subsession->sink = NULL;

    // Next, check whether *all* subsessions' streams have now been closed:
    MediaSession& session = subsession->parentSession();
    MediaSubsessionIterator iter(session);
    while ((subsession = iter.next()) != NULL)
    {
        if (subsession->sink != NULL) return; // this subsession is still active
    }

    // All subsessions' streams have now been closed, so shutdown the client:
    shutdownStream(pClient);
}

////////////////////////////////////////////////////////////////////////////
// subsessionByeHandler: Called when a RTCP "BYE" is received for a subsession
////////////////////////////////////////////////////////////////////////////
void subsessionByeHandler(void* clientData)
{
    LOCAL_DEBUG_0(__FUNCTION__,__LINE__);
    MediaSubsession* subsession = (MediaSubsession*)clientData;
    RTSPClient* pClient = (RTSPClient*)subsession->miscPtr;
    UsageEnvironment& env = pClient->envir(); // alias
    RAA_PARAMS* pParams = &((RTSPAsyncAdapter*)pClient)->params;
    if (pParams->byeCB)
        pParams->byeCB(pClient, pParams->cData);

    log_internal_write(L_INFO, __FILE__, __LINE__, RAA_I_RTCP_BYE_RECEIVED, pClient->url());

    // Now act as if the subsession had closed:
    postSubsessionPLAY(subsession);
}

////////////////////////////////////////////////////////////////////////////
// shutdownStream: Used to shut down and close a stream (including its "RTSPClient" object):
////////////////////////////////////////////////////////////////////////////
void shutdownStream(RTSPClient* pClient, int ec)
{
    LOCAL_DEBUG_0(__FUNCTION__,__LINE__);
    UsageEnvironment& env = pClient->envir(); // alias
    StreamClientState& scs = ((RTSPAsyncAdapter*)pClient)->scs; // alias

    // First, check whether any subsessions have still to be closed:
    if (scs.session != NULL)
    { 
        Boolean someSubsessionsWereActive = False;
        MediaSubsessionIterator iter(*scs.session);
        MediaSubsession* subsession;

        while ((subsession = iter.next()) != NULL)
        {
            if (subsession->sink != NULL)
            {
                Medium::close(subsession->sink);
                subsession->sink = NULL;
                someSubsessionsWereActive = True;
            }
        }
        if (someSubsessionsWereActive)
        {
            // Send a RTSP "TEARDOWN" command, to tell the server to shutdown the stream.
            // Don't bother handling the response to the "TEARDOWN".
            pClient->sendTeardownCommand(*scs.session, NULL);
        }
    }
    Medium::close(pClient);
    // Note that this will also cause this stream's "StreamClientState" structure to get reclaimed.
}

////////////////////////////////////////////////////////////////////////////
// TeardownStream: Used to teardown the session. shutdownStream() must be
//                 called subsequently for cleaning up resources.
////////////////////////////////////////////////////////////////////////////
void TeardownStream(RTSPClient* pClient, int ec)
{
    LOCAL_DEBUG_0(__FUNCTION__,__LINE__);
    UsageEnvironment& env = pClient->envir(); // alias
    StreamClientState& scs = ((RTSPAsyncAdapter*)pClient)->scs; // alias

    // First, check whether any subsessions have still to be closed:
    if (scs.session != NULL)
    { 
        Boolean someSubsessionsWereActive = False;
        MediaSubsessionIterator iter(*scs.session);
        MediaSubsession* subsession;

        while ((subsession = iter.next()) != NULL)
        {
            if (subsession->sink != NULL)
            {
                someSubsessionsWereActive = True;
                break;
            }
        }
        if (someSubsessionsWereActive)
        {
            // Send a RTSP "TEARDOWN" command, to tell the server to shutdown the stream.
            // Don't bother handling the response to the "TEARDOWN".
            LOCAL_DEBUG_1S(__FUNCTION__,__LINE__,"Sending request: TEARDOWN");
            pClient->sendTeardownCommand(*scs.session, NULL);
        }
    }
}

////////////////////////////////////////////////////////////////////////////
// Implementation of "RTSPAsyncAdapter"
////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////
// RTSPAsyncAdapter::createNew
////////////////////////////////////////////////////////////////////////////
RTSPAsyncAdapter* RTSPAsyncAdapter::createNew(UsageEnvironment& env,
                                              char const*       rtspURL,
					                          RAA_PARAMS*       pParams,
                                              int               verbosityLevel,
                                              char const*       appName,
                                              portNumBits       tunnelOverHTTPPortNum)
{
    LOCAL_DEBUG_0(__FUNCTION__,__LINE__);
    return new RTSPAsyncAdapter(env, rtspURL, pParams, verbosityLevel, appName, tunnelOverHTTPPortNum);
}

////////////////////////////////////////////////////////////////////////////
// constructor
////////////////////////////////////////////////////////////////////////////
RTSPAsyncAdapter::RTSPAsyncAdapter(UsageEnvironment& env, char const* rtspURL,
			     RAA_PARAMS* pParams, int verbosityLevel, char const* applicationName, portNumBits tunnelOverHTTPPortNum)
  : RTSPClient(env,rtspURL, verbosityLevel, applicationName, tunnelOverHTTPPortNum,-1)
{
    params = *pParams;
    LOCAL_DEBUG_2S(__FUNCTION__,__LINE__,"Connecting to:",rtspURL);
}

////////////////////////////////////////////////////////////////////////////
// destructor
////////////////////////////////////////////////////////////////////////////
RTSPAsyncAdapter::~RTSPAsyncAdapter()
{
}
////////////////////////////////////////////////////////////////////////////
// RTSPAsyncAdapter::RAA_Connect: Connect to URL and initiate stream
////////////////////////////////////////////////////////////////////////////
void RTSPAsyncAdapter::RAA_Connect( UsageEnvironment& env, const char* app, const char* url, RAA_PARAMS *pParams)
{
    LOCAL_DEBUG_0(__FUNCTION__,__LINE__);
    int vLevel;
    unsigned int rc = 0;
#ifdef _DEBUG
    vLevel = 1;
#else
    vLevel = 0;
#endif
    pParams->pEnv = &env;
    // Begin by creating a "RTSPClient" object.  Note that there is a separate "RTSPClient" object for each stream that we wish
    // to receive (even if more than stream uses the same "rtsp://" URL).
    *pParams->ppRTSPAdapter = (RTSPAsyncAdapter*)RTSPAsyncAdapter::createNew(env, url, pParams, vLevel, app);
    if (*pParams->ppRTSPAdapter == NULL)
    {
        LOCAL_DEBUG_2S(__FUNCTION__,__LINE__,"Failed to connect to:",pParams->url.c_str());
        log_internal_write(L_WARN, __FILE__, __LINE__, RAA_E_RTSP_CLIENT_CREATION_FAILED);
        return;
    }

    // Next, send a RTSP "DESCRIBE" command, to get a SDP description for the stream.
    // Note that this command - like all RTSP commands - is sent asynchronously; we do not block, waiting for a response.
    // Instead, the following function call returns immediately, and we handle the RTSP response later, from within the event loop:
    LOCAL_DEBUG_2S(__FUNCTION__,__LINE__,"Sending request: DESCRIBE",pParams->sessionMediumName);
    if (pParams->username.length())
        pParams->pAuth = new Authenticator(pParams->username.c_str(), pParams->pw.c_str());
    else
        pParams->pAuth = NULL;
    rc = (*pParams->ppRTSPAdapter)->sendDescribeCommand(postDESCRIBE, pParams->pAuth); 

    // All subsequent activity takes place within the event loop:
    env.taskScheduler().doEventLoop( pParams->pBlockingFlag ); // does not return

    // We never actually get to this line
}

////////////////////////////////////////////////////////////////////////////
// RTSPAsyncAdapter::RAA_Disconnect
////////////////////////////////////////////////////////////////////////////
void RTSPAsyncAdapter::RAA_Disconnect( RTSPAsyncAdapter* pClient)
{
    LOCAL_DEBUG_0(__FUNCTION__,__LINE__);
    if (pClient != NULL)
        TeardownStream( pClient, 0);
}

////////////////////////////////////////////////////////////////////////////
// RTSPAsyncAdapter::RAA_Shutdown
////////////////////////////////////////////////////////////////////////////
void RTSPAsyncAdapter::RAA_Shutdown( RTSPAsyncAdapter* pClient)
{
    LOCAL_DEBUG_0(__FUNCTION__,__LINE__);
    if (pClient != NULL)
        shutdownStream( pClient, 0);
}


// Implementation of "StreamClientState":

////////////////////////////////////////////////////////////////////////////
// constructor
////////////////////////////////////////////////////////////////////////////
StreamClientState::StreamClientState()
  : iter(NULL), session(NULL), subsession(NULL), streamTimerTask(NULL), duration(0.0)
{
    LOCAL_DEBUG_0(__FUNCTION__,__LINE__);
}

////////////////////////////////////////////////////////////////////////////
// destructor
////////////////////////////////////////////////////////////////////////////
StreamClientState::~StreamClientState()
{
    LOCAL_DEBUG_0(__FUNCTION__,__LINE__);
    if (iter)
    {
        delete iter;
        iter = NULL;
    }
    if (session != NULL)
    {
        // We also need to delete "session", and unschedule "streamTimerTask" (if set)
        UsageEnvironment& env = session->envir(); // alias

        env.taskScheduler().unscheduleDelayedTask(streamTimerTask);
        Medium::close(session);
    }
}


////////////////////////////////////////////////////////////////////////////
// Implementation of "RAA_Sink":
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// RAA_Sink::createNew
//    We need to receive the incoming data. Define the size of the buffer that we'll use:
////////////////////////////////////////////////////////////////////////////
RAA_Sink* RAA_Sink::createNew(UsageEnvironment& env,
                                MediaSubsession&  subsession,
                                unsigned char*    buf,
                                int               size,
                                AFTER_READING_CB  aftRead,
                                void*             cData,
                                char const*       streamId)
{
    LOCAL_DEBUG_1S(__FUNCTION__,__LINE__,"Creating sink: ", subsession.mediumName());
    return new RAA_Sink(env, subsession, buf, size, aftRead, cData, streamId);;
}

////////////////////////////////////////////////////////////////////////////
// constructor
////////////////////////////////////////////////////////////////////////////
RAA_Sink::RAA_Sink(UsageEnvironment&   env,
                     MediaSubsession&  subsession,
                     unsigned char    *buf,
                     int               size,
                     AFTER_READING_CB  aftRead,
                     void             *cData,
                     char const*       streamId)
                     : MediaSink(env), fSubsession(subsession)
{
    LOCAL_DEBUG_0(__FUNCTION__,__LINE__);
    fStreamId = strDup(streamId);
    fAfterReading = aftRead;
    fReceiveBuffer = buf;
    fClientData = cData;
    fSize = size;

    // Always increase Receive buffer to 2MB for video and to 100KB for audio
    if (subsession.rtpSource() != NULL)
    {
        int sock = subsession.rtpSource()->RTPgs()->socketNum();
        if ( _stricmp(subsession.mediumName(), "video") == 0)
        {
            LOCAL_DEBUG_SL(__FUNCTION__,__LINE__,"Setting video buffer size to", MAX_RTP_VIDEO_FRAME_SIZE);
            increaseReceiveBufferTo(env, sock, MAX_RTP_VIDEO_FRAME_SIZE);
        }
        else if ( _stricmp(subsession.mediumName(), "audio") == 0)
        {
            LOCAL_DEBUG_SL(__FUNCTION__,__LINE__,"Setting audio buffer size to", MAX_RTP_AUDIO_FRAME_SIZE);
            increaseReceiveBufferTo(env, sock, MAX_RTP_AUDIO_FRAME_SIZE);
        }
    }
}

////////////////////////////////////////////////////////////////////////////
// destructor
////////////////////////////////////////////////////////////////////////////
RAA_Sink::~RAA_Sink()
{
    LOCAL_DEBUG_2S(__FUNCTION__,__LINE__,"Closing sink:", fSubsession.mediumName());

    // No need to delete receive buffer. It belongs to the application,
    // which is expected to handle the recycling.
    delete[] fStreamId;
}

////////////////////////////////////////////////////////////////////////////
// RAA_Sink::afterGettingFrame
////////////////////////////////////////////////////////////////////////////
void RAA_Sink::afterGettingFrame(void* clientData, unsigned frameSize, unsigned numTruncatedBytes,
				  struct timeval presentationTime, unsigned durationInMicroseconds)
{
    RAA_Sink* sink = (RAA_Sink*)clientData;
    LOCAL_DEBUG_2S(__FUNCTION__,__LINE__,"received: ", sink->fSubsession.mediumName());
    sink->afterGettingFrame(frameSize, numTruncatedBytes, presentationTime, durationInMicroseconds);
}

////////////////////////////////////////////////////////////////////////////
// RAA_Sink::afterGettingFrame
////////////////////////////////////////////////////////////////////////////
void RAA_Sink::afterGettingFrame(unsigned frameSize, unsigned numTruncatedBytes,
				  struct timeval presentationTime, unsigned durationInMicroseconds) 
{
    LOCAL_DEBUG_0(__FUNCTION__,__LINE__);

    // We've just received a frame of data.  Inform the application about it.
    if (fAfterReading)
    {
        LOCAL_DEBUG_2S(__FUNCTION__,__LINE__,"sending frame: ", fSubsession.mediumName());
        bool hasBeenSynched = fSubsession.rtpSource() != NULL && fSubsession.rtpSource()->hasBeenSynchronizedUsingRTCP();
        (*fAfterReading)(fClientData, frameSize, numTruncatedBytes, presentationTime, durationInMicroseconds, hasBeenSynched);
    }
  
    // Then continue, to request the next frame of data:
    continuePlaying();
}

////////////////////////////////////////////////////////////////////////////
// RAA_Sink::continuePlaying
////////////////////////////////////////////////////////////////////////////
Boolean RAA_Sink::continuePlaying()
{
    if (fSource == NULL) return False; // sanity check (should not happen)

    LOCAL_DEBUG_1S(__FUNCTION__,__LINE__,"Get next frame");

    // Request the next frame of data from our input source.
    // "afterGettingFrame()" will get called later, when it arrives.
    fSource->getNextFrame(fReceiveBuffer, fSize,
                          afterGettingFrame, this,
                          onSourceClosure, this);
    return True;
}

////////////////////////////////////////////////////////////////////////////
// RAA_NetStats Implementation
////////////////////////////////////////////////////////////////////////////

static void periodicStats(void* cData); // forward

////////////////////////////////////////////////////////////////////////////
// constructor
////////////////////////////////////////////////////////////////////////////
RAA_NetStats::RAA_NetStats(struct timeval const& startTime, RTPSource* src)
    : fSource(src)
    , fNext(NULL)
    , kbits_per_second_min(1e20)
    , kbits_per_second_max(0)
    , kBytesTotal(0.0)
    , packet_loss_fraction_min(1.0)
    , packet_loss_fraction_max(0.0)
    , totNumPacketsReceived(0)
    , totNumPacketsExpected(0)
{
    LOCAL_DEBUG_0(__FUNCTION__,__LINE__);
    measurementEndTime = measurementStartTime = startTime;

    RTPReceptionStatsDB::Iterator statsIter(src->receptionStatsDB());

    // Assume that there's only one SSRC source (usually the case):
    RTPReceptionStats* stats = statsIter.next(True);
    if (stats != NULL)
    {
        kBytesTotal = stats->totNumKBytesReceived();
        totNumPacketsReceived = stats->totNumPacketsReceived();
        totNumPacketsExpected = stats->totNumPacketsExpected();
    }
}

////////////////////////////////////////////////////////////////////////////
// destructor
////////////////////////////////////////////////////////////////////////////
RAA_NetStats::~RAA_NetStats()
{
    LOCAL_DEBUG_0(__FUNCTION__,__LINE__);
    delete fNext;
}

////////////////////////////////////////////////////////////////////////////
// RAA_NetStats::periodicStats
////////////////////////////////////////////////////////////////////////////
void RAA_NetStats::periodicStats(struct timeval const& timeNow, void* cData)
{
    LOCAL_DEBUG_0(__FUNCTION__,__LINE__);
    RAA_PARAMS *pParams = (RAA_PARAMS*)cData;
    unsigned secsDiff = timeNow.tv_sec - measurementEndTime.tv_sec;
    int usecsDiff = timeNow.tv_usec - measurementEndTime.tv_usec;
    double timeDiff = secsDiff + usecsDiff/1000000.0;
    measurementEndTime = timeNow;

    RTPReceptionStatsDB::Iterator statsIter(fSource->receptionStatsDB());

    // Assume that there's only one SSRC source (usually the case):
    RTPReceptionStats* stats = statsIter.next(True);
    if (stats != NULL)
    {
        double kBytesTotalNow = stats->totNumKBytesReceived();
        double kBytesDeltaNow = kBytesTotalNow - kBytesTotal;
        kBytesTotal = kBytesTotalNow;

        double kbpsNow = timeDiff == 0.0 ? 0.0 : 8*kBytesDeltaNow/timeDiff;
        if (kbpsNow < 0.0) kbpsNow = 0.0; // in case of roundoff error
        if (kbpsNow < kbits_per_second_min) kbits_per_second_min = kbpsNow;
        if (kbpsNow > kbits_per_second_max) kbits_per_second_max = kbpsNow;

        unsigned totReceivedNow = stats->totNumPacketsReceived();
        unsigned totExpectedNow = stats->totNumPacketsExpected();
        unsigned deltaReceivedNow = totReceivedNow - totNumPacketsReceived;
        unsigned deltaExpectedNow = totExpectedNow - totNumPacketsExpected;
        totNumPacketsReceived = totReceivedNow;
        totNumPacketsExpected = totExpectedNow;
        pParams->netStats.m_strmPktLoss = totExpectedNow - totReceivedNow;
        pParams->netStats.m_strmJitter  = stats->jitter();

        if ( pParams->statsCB)
            pParams->statsCB( &pParams->netStats, pParams->cData );
    }
}

////////////////////////////////////////////////////////////////////////////
// scheduleNextStats
////////////////////////////////////////////////////////////////////////////
static void scheduleNextStats(RAA_PARAMS *pParams)
{
    pParams->nextStatsMeasurementUSecs += pParams->statsMeasurementIntervalMS * 1000;
    LOCAL_DEBUG_1L(__FUNCTION__,__LINE__,pParams->nextStatsMeasurementUSecs);
    struct timeval timeNow;
    gettimeofday(&timeNow, NULL);
    unsigned timeNowUSecs = timeNow.tv_sec*1000000 + timeNow.tv_usec;
    int usecsToDelay = pParams->nextStatsMeasurementUSecs - timeNowUSecs;

    pParams->statsMeasurementTimerTask = pParams->pEnv->taskScheduler().scheduleDelayedTask(
    usecsToDelay, (TaskFunc*)periodicStats, (void*)pParams);
}

////////////////////////////////////////////////////////////////////////////
// periodicStats
////////////////////////////////////////////////////////////////////////////
static void periodicStats(void* cData)
{
    LOCAL_DEBUG_0(__FUNCTION__,__LINE__);
    RAA_PARAMS* pParams = (RAA_PARAMS*)cData;
    struct timeval timeNow;
    gettimeofday(&timeNow, NULL);

    for (RAA_NetStats* pStats = pParams->pStatsHead;
    pStats != NULL; pStats = pStats->fNext)
    {
        pStats->periodicStats(timeNow, pParams);
    }

    // Do this again later:
    scheduleNextStats(pParams);
}

////////////////////////////////////////////////////////////////////////////
// periodicStats
////////////////////////////////////////////////////////////////////////////
static void beginStatsMeasurement(void* cData, MediaSession* session)
{
    LOCAL_DEBUG_0(__FUNCTION__,__LINE__);
    RAA_PARAMS *pParams = (RAA_PARAMS*)cData;

    // Set up a measurement record for each active subsession
    struct timeval startTime;
    gettimeofday(&startTime, NULL);
    pParams->nextStatsMeasurementUSecs = startTime.tv_sec*1000000 + startTime.tv_usec;
    RAA_NetStats* pStatsTail = NULL;
    MediaSubsessionIterator iter(*session);
    MediaSubsession* subsession;
    while ((subsession = iter.next()) != NULL)
    {
        RTPSource* src = subsession->rtpSource();
        if (src == NULL) continue;

        RAA_NetStats* pStats = new RAA_NetStats(startTime, src);
        if (pStats != NULL) pParams->pStatsHead = pStats;
        if (pStatsTail != NULL) pStatsTail->fNext = pStats;
        pStatsTail  = pStats;
    }

    // Then schedule the first of the periodic measurements:
    scheduleNextStats(pParams);
}
