////////////////////////////////////////////////////////////////////////////
// 2015 UTC Buildings & Industrial Systems. All rights reserved.
// This software contains information which represents trade secrets of UTC Buildings & Industrial Systems and may not be copied or disclosed
// to others except as provided in the license with UTC Buildings & Industrial Systems.

// Code in this module is based on the demo application, testRTSPCLient, copied from Live555 website.
// The code underwent significant modifications to suit UTC needs.
// For more details about testRTSPClient demo application and OpenRTSP, see http://www.live555.com/openRTSP/
////////////////////////////////////////////////////////////////////////////
// RTSP Async Adapter - Include file
////////////////////////////////////////////////////////////////////////////
#pragma once

#ifndef RTSP_ASYNC_ADAPTER_H
#define RTSP_ASYNC_ADAPTER_H

#include <string>
#include "liveMedia.hh"
#include "BasicUsageEnvironment.hh"
#include "GroupsockHelper.hh"
#include "../../utils/dbg_out.h"

#define MAX_RTP_VIDEO_FRAME_SIZE (2 * 1024 * 1024)
#define MAX_RTP_AUDIO_FRAME_SIZE (100 * 1024)
#define RAA_PAYLOAD_TYPE_VIDEO 1
#define RAA_PAYLOAD_TYPE_AUDIO 2
#define RAA_PAYLOAD_TYPE_UNKNOWN 0

////////////////////////////////////////////////////////////////////////////
// RTSP Async Adapter - Error codes
////////////////////////////////////////////////////////////////////////////
#define RAA_ERROR_UNAUTHORIZED    401

////////////////////////////////////////////////////////////////////////////
// RTSP Async Adapter - Callbacks from Adapter into Prism driver
////////////////////////////////////////////////////////////////////////////

typedef void (*DESCRIBE_CB)(RTSPClient* pClient, void* cData, int rc, char* sdp);
typedef void (*META_CB)(RTSPClient* pClient, void* cData, int rc, char* codec, char* sParams);
typedef void (*AFTER_READING_CB)(void* cData, unsigned int frameSize, unsigned int truncBytes, struct timeval presTime, unsigned int duration, bool hasBeenSynched);
typedef void (*BYE_HANDLER_CB)(RTSPClient* pClient, void* cData);
typedef void (*SESSION_STATUS_CB)(RTSPClient* pClient, void* cData);
typedef void (*SDP_ERROR_CB)(RTSPClient* pClient, int rc, void* pParams);
typedef void (*STATS_CB)(CPrismSessionNetStats* pStats, void* cData);

class RTSPAsyncAdapter;     // forward
class RAA_NetStats;         // forward

////////////////////////////////////////////////////////////////////////////
// RTSP Async Adapter - Information shared between Adapter and Prism driver
////////////////////////////////////////////////////////////////////////////
typedef struct RAA_Params
{
    DESCRIBE_CB           descCB;
    META_CB               metaCB;
    AFTER_READING_CB      frameCB;
    BYE_HANDLER_CB        byeCB;
    SESSION_STATUS_CB     subsessStatusCB;
    SDP_ERROR_CB          sdpErrorCB;
    STATS_CB              statsCB;

    char                 *frameBuf;
    int                   frameSize;
    UsageEnvironment     *pEnv;
    RTSPAsyncAdapter    **ppRTSPAdapter;
    std::string           url;
    std::string           username;
    std::string           pw;
    std::string           domain;
    void                 *cData;
    int                   payloadType;
    char                 *pBlockingFlag;
    CPrismSessionNetStats netStats;
    unsigned int          lastPktLoss;
    RAA_NetStats         *pStatsHead;
    unsigned int          nextStatsMeasurementUSecs;
    unsigned int          statsMeasurementIntervalMS;
    TaskToken             statsMeasurementTimerTask;
    char                 *sessionMediumName;
    Authenticator        *pAuth;
    bool                 *pReGetSDP;
    int                   playbackspeed;
} RAA_PARAMS;


////////////////////////////////////////////////////////////////////////////
// Define a class to hold per-stream state that we maintain throughout each stream's lifetime:
////////////////////////////////////////////////////////////////////////////
class StreamClientState
{
public:
    StreamClientState();
    virtual ~StreamClientState();

public:
    MediaSubsessionIterator* iter;
    MediaSession*            session;
    MediaSubsession*         subsession;
    TaskToken                streamTimerTask;
    double                   duration;
};

////////////////////////////////////////////////////////////////////////////
// RTSP Async Adapter class definition
////////////////////////////////////////////////////////////////////////////
class RTSPAsyncAdapter: public RTSPClient
{
public:
    static RTSPAsyncAdapter* createNew(UsageEnvironment& env,
                                       char const*       rtspURL,
				                       RAA_PARAMS*       pParams,
                                       int               verbosityLevel = 0,
				                       char const*       appName = NULL,
				                       portNumBits       tunnelOverHTTPPortNum = 0);

    static void RAA_Connect( UsageEnvironment& env,
                             const char*       app, 
                             const char*       url,
                             RAA_PARAMS*       pParams);

    static void RAA_Disconnect( RTSPAsyncAdapter* pAdapter);
    static void RAA_Shutdown( RTSPAsyncAdapter* pAdapter );

protected:
    RTSPAsyncAdapter(UsageEnvironment& env,
                     char const*       rtspURL,
		             RAA_PARAMS*       pParams,
                     int               verbosityLevel,
                     char const*       appName,
                     portNumBits       tunnelOverHTTPPortNum);

    // called only by createNew();
    virtual ~RTSPAsyncAdapter();

public:
    StreamClientState scs;
    RAA_PARAMS params;
};
////////////////////////////////////////////////////////////////////////////
// RAA_Sink class definition
////////////////////////////////////////////////////////////////////////////
class RAA_Sink: public MediaSink
{
public:
  static RAA_Sink* createNew(UsageEnvironment& env,
			      MediaSubsession& subsession, // identifies the kind of data that's being received
                  unsigned char    *buf,
                  int               size,
			      AFTER_READING_CB aftRead,
                  void             *cData,
                  char const* streamId = NULL); // identifies the stream itself (optional)

private:
    RAA_Sink(UsageEnvironment& env,
            MediaSubsession&  subsession,
            unsigned char    *buf,
            int               size,
            AFTER_READING_CB  aftRead,
            void             *cData,
            char const*       streamId);
    // called only by "createNew()"
    virtual ~RAA_Sink();

    static void afterGettingFrame(void* clientData, unsigned frameSize,
                                unsigned numTruncatedBytes,
				struct timeval presentationTime,
                                unsigned durationInMicroseconds);
    void afterGettingFrame(unsigned frameSize, unsigned numTruncatedBytes,
			 struct timeval presentationTime, unsigned durationInMicroseconds);

private:
    // redefined virtual functions:
    virtual Boolean continuePlaying();

private:
    u_int8_t*        fReceiveBuffer;
    int              fSize;
    MediaSubsession& fSubsession;
    char*            fStreamId;
    AFTER_READING_CB fAfterReading;
    void*            fClientData;
};
////////////////////////////////////////////////////////////////////////////
// RTSP Async Adapter - Network Statistics Information shared between
// Adapter and Live555 Session
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// RAA_NetStats class definition
////////////////////////////////////////////////////////////////////////////
class RAA_NetStats
{
public:
    RAA_NetStats(struct timeval const& startTime, RTPSource* src);
    virtual ~RAA_NetStats();

    void periodicStats(struct timeval const& timeNow, void *cData);

public:
    RTPSource*    fSource;
    RAA_NetStats* fNext;

public:
    struct timeval measurementStartTime, measurementEndTime;
    double         kbits_per_second_min, kbits_per_second_max;
    double         kBytesTotal;
    double         packet_loss_fraction_min, packet_loss_fraction_max;
    unsigned       totNumPacketsReceived, totNumPacketsExpected;
};

#endif //RTSP_ASYNC_ADAPTER_H
